import time
import dash
from dash import dcc, html, Input, Output, State
import dash_bootstrap_components as dbc
import os
import threading

from cvasl_gui.app import app
from cvasl_gui.tabs.job_list import run_job, get_job_status
from cvasl_gui import data_store
from cvasl_gui.components.directory_input import create_directory_input
from cvasl_gui.components.data_table import create_data_table
from cvasl_gui.components.feature_compare import create_feature_compare

from cvasl_gui.tabs.parameters import parameters

# Folder where job output files are stored
WORKING_DIR = os.getenv("CVASL_WORKING_DIRECTORY", ".")
INPUT_DIR = os.path.join(WORKING_DIR, 'data')
JOBS_DIR = os.path.join(WORKING_DIR, 'jobs')

ALGORITHM_PARAMS = {
    "neurocombat": {
        "label": "NeuroCombat",
        "parameters": ["features_to_harmonize", "discrete_covariates", "continuous_covariates", "site_indicator", "patient_identifier", "empirical_bayes", "mean_only", "parametric" ]
    },
    "neuroharmonize": {
        "label": "NeuroHarmonize",
        "parameters": ["features_to_harmonize", "covariates", "smooth_terms", "site_indicator", "empirical_bayes" ]
    },
    "covbat": {
        "label": "CovBat",
        "parameters": ["features_to_harmonize", "covariates", "site_indicator", "patient_identifier", "numerical_covariates", "empirical_bayes" ]
    },
    "nestedcombat": {
        "label": "OPNested ComBat",
        "parameters": ["features_to_harmonize", "batch_list_harmonisations", "site_indicator", "discrete_covariates", "continuous_covariates", "intermediate_results_path", "patient_identifier", "return_extended", "use_gmm" ]
    },
    "autocombat": {
        "label": "AutoComBat",
        "parameters": ["data_subset", "features_to_harmonize", "site_indicator", "discrete_covariates", "continuous_covariates", "discrete_cluster_features", "continuous_cluster_features", "metric", "features_reduction", "feature_reduction_dimensions", "empirical_bayes" ]
    },
    "relief": {
        "label": "RELIEF",
        "parameters": ["features_to_harmonize", "covariates", "patient_identifier", "intermediate_results_path" ]
    }
}



def get_dataframe_columns():
    df = data_store.all_data['harmonization']
    if df is None:
        return []
    return df.columns


def create_tab_harmonization():
    return html.Div([
        dbc.Accordion([
            dbc.AccordionItem([create_directory_input('harmonization')],
                title="Select new input data"),
            dbc.AccordionItem([create_data_table('harmonization')],
                title="Inspect data"),
            dbc.AccordionItem([create_feature_compare()],
                title="Feature comparison"),
            dbc.AccordionItem(create_harmonization_parameters(),
                title="Harmonization"),
            dcc.Store(id="harmonization-job-id", data=None),
        ], always_open=True)
    ])


def create_parameter_component(parameter_data, width=3):
    id = parameter_data[0]
    label_text = parameter_data[1]["label"]
    parameter_type = parameter_data[1]["type"]
    description = parameter_data[1].get("description", "")
    default_value = parameter_data[1].get("default", None)

    # Depending on the type, create the appropriate input component
    parameter_control = None
    match parameter_type:
        case "feature-list-multi":
            parameter_control = dcc.Dropdown(id=f"{id}-dropdown", options=[],
                                            multi=True, placeholder="select features...",
                                            value=default_value),
        case "feature-list-single":
            parameter_control = dcc.Dropdown(id=f"{id}-dropdown", options=[],
                                            multi=False, placeholder="select feature...",
                                            value=default_value),
        case "selection":
            options = parameter_data[1].get("options", [])
            parameter_control = dcc.Dropdown(id=f"{id}-dropdown", 
                                            options=[{"label": opt, "value": opt} for opt in options],
                                            multi=False, placeholder="select option...",
                                            value=default_value),
        case "str":
            parameter_control = dbc.Input(id=f"{id}-input", type="text",
                                         placeholder="enter value...",
                                         value=default_value),
        case "int":
            parameter_control = dbc.Input(id=f"{id}-input", type="number",
                                         placeholder="enter integer...",
                                         value=default_value),
        case "bool":
            bool_default = default_value if default_value is not None else False
            parameter_control = dbc.Checkbox(id=f"{id}-checkbox", value=bool_default,
                                             style={"marginTop": "6px"}),

    return dbc.Row([
        dbc.Col(
            html.Div([
                html.Label(
                    label_text,
                    style={"marginTop": "6px", "marginRight": "5px", "marginBottom": "0"}
                ),
                html.Span(
                    "i",
                    id=f"{id}-target",
                    style={
                        "cursor": "pointer",
                        "fontFamily": "serif",
                        "fontWeight": "bold",
                        "fontStyle": "italic",
                        "border": "2px solid #97b0d4",
                        "borderRadius": "50%",
                        "display": "inline-flex",
                        "alignItems": "center",
                        "justifyContent": "center",
                        "width": "18px",
                        "height": "18px",
                        "fontSize": "12px",
                        "lineHeight": "1",
                        "textAlign": "center",
                        "color": "#97b0d4",
                        "verticalAlign": "middle"
                    }
                ),
                dbc.Tooltip(
                    description,
                    target=f"{id}-target",
                    placement="right",
                    trigger="hover"
                )
            ], style={"display": "flex", "alignItems": "center", "gap": "5px"}
            ), width=width
        ),
        dbc.Col(parameter_control),
    ], className="mb-3", id=f"row-{id}")



def create_harmonization_parameters():

    return [
        dbc.Row([
            dbc.Col(html.Label("Algorithm", style={"marginTop": "6px"}), width=3),
            dbc.Col(
                dcc.Dropdown(
                    id="algorithm-dropdown",
                    options=[
                        {"label": v["label"], "value": k} for k, v in ALGORITHM_PARAMS.items()
                    ],
                    value="neurocombat",
                    clearable=False,
                ),
            ),
        ], className="mb-3", id="row-algorithm"),

        # Instantiate a parameter component for each parameter,
        # they will be dynamically shown/hidden based on the selected algorithm
        *[create_parameter_component(parameter_data) for parameter_data in parameters.items()],

        dbc.Row([
            dbc.Col(html.Label("Label", style={"marginTop": "6px"}), width=3),
            dbc.Col(dbc.Input(id="label-input", type="text",
                placeholder="Enter label...", value="neurocombat"),
            ),
        ], className="mb-3", id="row-label-input"),

        # Row for button and status
        dbc.Row([
            dbc.Col(html.Label("", style={"marginTop": "6px"}), width=3),
            dbc.Col(
                html.Div([
                    html.Button("Start harmonization", id="start-button", n_clicks=0),
                    html.Span("Status: ", style={"marginLeft": "10px"}),
                    html.Span(id="harmonization-job-status", children="Idle"),
                    dcc.Interval(id="harmonization-status-interval", interval=1000, n_intervals=0, disabled=True)
                ]),
            ),
        ], className="mb-3"),
    ]


@app.callback(
    [ Output(f"row-{id}", "style") for id in parameters.keys() ],
    Input("algorithm-dropdown", "value"),
)
def toggle_rows(algorithm):
    """Show/hide parameter rows based on the selected algorithm and hide property"""
    active_parameters = ALGORITHM_PARAMS.get(algorithm, {}).get("parameters", [])
    
    styles = []
    for param_id in parameters.keys():
        # Check if parameter should be hidden
        should_hide = parameters[param_id].get("hide", False)
        
        # Show if it's active for the algorithm and not marked as hidden
        if param_id in active_parameters and not should_hide:
            styles.append({"display": "flex"})
        else:
            styles.append({"display": "none"})
    
    return styles


@app.callback(
    Output("label-input", "value"),
    Input("algorithm-dropdown", "value"),
)
def update_label_from_algorithm(algorithm):
    """Update the label input to match the selected algorithm name"""
    return algorithm.lower() if algorithm else "harmonized"


#
#  Populate the dropdowns with columns from the data table
#

@app.callback(
    Output("data_subset-dropdown", "options"),
    Input({'type': 'data-table', 'index': 'harmonization'}, "data"),
    prevent_initial_call=True
)
def update_data_subset_dropdown(data):
    if not data:
        return []
    return [{"label": col, "value": col} for col in data[0].keys()]

@app.callback(
    Output("features_to_harmonize-dropdown", "options"),
    Input({'type': 'data-table', 'index': 'harmonization'}, "data"),
    prevent_initial_call=True
)
def update_feature_dropdown(data):
    if not data: # Check if data is empty
        return []
    return [{"label": col, "value": col} for col in data[0].keys()]

@app.callback(
    Output("discrete_covariates-dropdown", "options"),
    Output("discrete_covariates-dropdown", "value"),
    Input({'type': 'data-table', 'index': 'harmonization'}, "data"),
    prevent_initial_call=True
)
def update_discrete_covariate_dropdown(data):
    if not data:
        return [], None
    options = [{"label": col, "value": col} for col in data[0].keys()]
    default_value = ["Sex"] if "Sex" in data[0].keys() else None
    return options, default_value

@app.callback(
    Output("continuous_covariates-dropdown", "options"),
    Output("continuous_covariates-dropdown", "value"),
    Input({'type': 'data-table', 'index': 'harmonization'}, "data"),
    prevent_initial_call=True
)
def update_continuous_covariate_dropdown(data):
    if not data:
        return [], None
    options = [{"label": col, "value": col} for col in data[0].keys()]
    default_value = ["Age"] if "Age" in data[0].keys() else None
    return options, default_value

@app.callback(
    Output("site_indicator-dropdown", "options"),
    Output("site_indicator-dropdown", "value"),
    Input({'type': 'data-table', 'index': 'harmonization'}, "data"),
    prevent_initial_call=True
)
def update_site_indicator_dropdown(data):
    if not data:
        return [], None
    options = [{"label": col, "value": col} for col in data[0].keys()]
    default_value = "Site" if "Site" in data[0].keys() else None
    return options, default_value

@app.callback(
    Output("patient_identifier-dropdown", "options"),
    Output("patient_identifier-dropdown", "value"),
    Input({'type': 'data-table', 'index': 'harmonization'}, "data"),
    prevent_initial_call=True
)
def update_patient_identifier_dropdown(data):
    if not data:
        return [], None
    options = [{"label": col, "value": col} for col in data[0].keys()]
    default_value = "participant_id" if "participant_id" in data[0].keys() else None
    return options, default_value

@app.callback(
    Output("covariates-dropdown", "options"),
    Input({'type': 'data-table', 'index': 'harmonization'}, "data"),
    prevent_initial_call=True
)
def update_covariates_dropdown(data):
    if not data:
        return []
    return [{"label": col, "value": col} for col in data[0].keys()]

@app.callback(
    Output("batch_list_harmonisations-dropdown", "options"),
    Input({'type': 'data-table', 'index': 'harmonization'}, "data"),
    prevent_initial_call=True
)
def update_batch_list_harmonisations_dropdown(data):
    if not data:
        return []
    return [{"label": col, "value": col} for col in data[0].keys()]

@app.callback(
    Output("numerical_covariates-dropdown", "options"),
    Input({'type': 'data-table', 'index': 'harmonization'}, "data"),
    prevent_initial_call=True
)
def update_numerical_covariates_dropdown(data):
    if not data:
        return []
    return [{"label": col, "value": col} for col in data[0].keys()]

@app.callback(
    Output("smooth_terms-dropdown", "options"),
    Input({'type': 'data-table', 'index': 'harmonization'}, "data"),
    prevent_initial_call=True
)
def update_smooth_terms_dropdown(data):
    if not data:
        return []
    return [{"label": col, "value": col} for col in data[0].keys()]

@app.callback(
    Output("discrete_cluster_features-dropdown", "options"),
    Input({'type': 'data-table', 'index': 'harmonization'}, "data"),
    prevent_initial_call=True
)
def update_discrete_cluster_features_dropdown(data):
    if not data:
        return []
    return [{"label": col, "value": col} for col in data[0].keys()]

@app.callback(
    Output("continuous_cluster_features-dropdown", "options"),
    Input({'type': 'data-table', 'index': 'harmonization'}, "data"),
    prevent_initial_call=True
)
def update_continuous_cluster_features_dropdown(data):
    if not data:
        return []
    return [{"label": col, "value": col} for col in data[0].keys()]


@app.callback(
    Output("harmonization-job-id", "data"),
    Output("harmonization-status-interval", "disabled", allow_duplicate=True),
    Input("start-button", "n_clicks"),
    State("algorithm-dropdown", "value"),
    State("label-input", "value"),
    # Add all parameter controls as State
    State("data_subset-dropdown", "value"),
    State("features_to_harmonize-dropdown", "value"),
    State("batch_list_harmonisations-dropdown", "value"),
    State("covariates-dropdown", "value"),
    State("discrete_covariates-dropdown", "value"),
    State("continuous_covariates-dropdown", "value"),
    State("numerical_covariates-dropdown", "value"),
    State("smooth_terms-dropdown", "value"),
    State("discrete_cluster_features-dropdown", "value"),
    State("continuous_cluster_features-dropdown", "value"),
    State("metric-dropdown", "value"),
    State("features_reduction-dropdown", "value"),
    State("feature_reduction_dimensions-input", "value"),
    State("empirical_bayes-checkbox", "value"),
    State("mean_only-checkbox", "value"),
    State("parametric-checkbox", "value"),
    State("return_extended-checkbox", "value"),
    State("use_gmm-checkbox", "value"),
    State("site_indicator-dropdown", "value"),
    State("patient_identifier-dropdown", "value"),
    State("intermediate_results_path-input", "value"),
    prevent_initial_call=True
)
def start_job(n_clicks, algorithm, label, 
              data_subset, features_to_harmonize, batch_list_harmonisations, covariates,
              discrete_covariates, continuous_covariates, numerical_covariates, smooth_terms,
              discrete_cluster_features, continuous_cluster_features, metric, features_reduction,
              feature_reduction_dimensions, empirical_bayes, mean_only, parametric, 
              return_extended, use_gmm, site_indicator, patient_identifier, intermediate_results_path):
    
    if not features_to_harmonize:
        return dash.no_update, True
    
    # Map parameter names to their values
    param_values = {
        "data_subset": data_subset,
        "features_to_harmonize": features_to_harmonize,
        "batch_list_harmonisations": batch_list_harmonisations,
        "covariates": covariates,
        "discrete_covariates": discrete_covariates,
        "continuous_covariates": continuous_covariates,
        "numerical_covariates": numerical_covariates,
        "smooth_terms": smooth_terms,
        "discrete_cluster_features": discrete_cluster_features,
        "continuous_cluster_features": continuous_cluster_features,
        "metric": metric,
        "features_reduction": features_reduction,
        "feature_reduction_dimensions": feature_reduction_dimensions,
        "empirical_bayes": empirical_bayes,
        "mean_only": mean_only,
        "parametric": parametric,
        "return_extended": return_extended,
        "use_gmm": use_gmm,
        "site_indicator": site_indicator,
        "patient_identifier": patient_identifier,
        "intermediate_results_path": intermediate_results_path,
    }

    if algorithm == "autocombat":
        # Create a data_subset that consists of the features, covariates and site indicators concatenated
        combined_subset = set(features_to_harmonize or []) | set(discrete_covariates or []) | \
                set(continuous_covariates or []) | set(discrete_cluster_features or []) | \
                set(continuous_cluster_features or []) | set([site_indicator] if site_indicator else [])
        param_values["data_subset"] = list(combined_subset)

    # Get active parameters for the selected algorithm
    algorithm_parameters = ALGORITHM_PARAMS.get(algorithm, {}).get("parameters", [])

    # Build parameters dict with only active parameters that have values
    job_parameters = {}
    for param in algorithm_parameters:
        if param in param_values:
            value = param_values[param]
            param_type = parameters[param].get("type")
            if param_type == "feature-list-multi" and value is None:
                value = []
            if value is None:
                continue  # Skip parameters without a value
            # Beyond this point, None is considered a valid value
            if param_type == "selection" and value == "None":
                value = None
            job_parameters[param] = value

    job_arguments = {
        "type": "harmonization",
        "algorithm": algorithm,
        "input_paths": data_store.input_files['harmonization'],
        "input_sites": data_store.input_sites['harmonization'],
        "parameters": job_parameters,
        "label": label,
    }

    # Generate job_id and start the job in a separate thread
    job_id = str(int(time.time())) + "_" + algorithm
    threading.Thread(target=run_job, args=(job_arguments, job_id, True), daemon=True).start()

    return job_id, False  # enable the interval


@app.callback(
    Output("harmonization-job-status", "children"),
    Output("harmonization-status-interval", "disabled", allow_duplicate=True),
    Input("harmonization-status-interval", "n_intervals"),
    State("harmonization-job-id", "data"),
    prevent_initial_call=True,
)
def update_job_status(n, job_id):
    if not job_id:
        return "", True

    status = get_job_status(job_id)

    if status.lower() in ("completed", "failed", "cancelled"):
        return status, True  # Stop interval
    return status, False




selected_directory = None
all_data = {
  'harmonization': None,
  'prediction-training': None,
  'prediction-testing': None,
}
input_files = {
  'harmonization': None,
  'prediction-training': None,
  'prediction-testing': None,
}
input_sites = {
  'harmonization': None,
  'prediction-training': None,
  'prediction-testing': None,
}



import dash
from dash import dcc, html, Input, Output, State, ctx, MATCH, ALL
import os
import sys
import time
import subprocess
import json
import signal
import traceback

from cvasl_gui.app import app


# Folder where job output files are stored
WORKING_DIR = os.getenv("CVASL_WORKING_DIRECTORY", ".")
INPUT_DIR = os.path.join(WORKING_DIR, 'data')
JOBS_DIR = os.path.join(WORKING_DIR, 'jobs')



def create_job_list():
    return html.Div([
        html.Div(id="job-status"),
        dcc.Loading(
            id="loading-1",
            type="circle",
            children=html.Div(id="loading-output-job-status")
        ),
        dcc.Interval(id="interval-check", interval=3000, n_intervals=0),
        dcc.Download(id="download-data")
    ])

def run_job(job_arguments: dict, job_id: str, is_harmonization: bool = True):
    """Function to start the harmonization job"""

    # Create a unique folder for the job
    job_folder = os.path.join(JOBS_DIR, job_id)
    os.makedirs(job_folder, exist_ok=True)

    # Set the intermediate results path to the job's directory
    if "intermediate_results_path" in job_arguments["parameters"]:
        job_arguments["parameters"]["intermediate_results_path"] = job_folder

    # Save job arguments
    with open(os.path.join(job_folder, "job_arguments.json"), "w") as f:
        json.dump(job_arguments, f)

    # Start the job
    print("Starting job", job_id)
    if is_harmonization:
        script_path = os.path.join(os.path.dirname(__file__), "..", "jobs", "harmonization_job.py")
    else:
        script_path = os.path.join(os.path.dirname(__file__), "..", "jobs", "prediction_job.py")
    
    # Start process with output redirected to files
    try:
        print(f"Running script: {script_path} with job ID: {job_id}")
        
        # Create output files for stdout and stderr
        stdout_file = os.path.join(job_folder, "stdout.log")
        stderr_file = os.path.join(job_folder, "stderr.log")
        
        with open(stdout_file, "w") as stdout_f, open(stderr_file, "w") as stderr_f:
            process = subprocess.Popen(
                [sys.executable, script_path, job_id],
                stdout=stdout_f,
                stderr=stderr_f,
                text=True
            )
        print(f"Started process with PID: {process.pid}")
        
        # Give the process a moment to start and check for immediate failures
        time.sleep(1)
        poll_result = process.poll()

        print(f"Process poll result: {poll_result}")
        
        if poll_result is not None and poll_result != 0:
            # Process has already exited with non-zero code - this indicates a startup error
            error_msg = f"Process failed to start (exit code {poll_result})\n"
            
            # Read stderr and stdout from files if they exist
            if os.path.exists(stderr_file):
                with open(stderr_file, "r") as f:
                    stderr_content = f.read()
                    if stderr_content:
                        error_msg += f"Error output:\n{stderr_content}\n"
            
            if os.path.exists(stdout_file):
                with open(stdout_file, "r") as f:
                    stdout_content = f.read()
                    if stdout_content:
                        error_msg += f"Standard output:\n{stdout_content}\n"
            
            # Write error to log file
            error_log_path = os.path.join(job_folder, "error.log")
            with open(error_log_path, "w") as f:
                f.write(error_msg)
            
            # Write failed status
            status_file = os.path.join(job_folder, "job_status")
            with open(status_file, "w") as f:
                f.write("failed")
            
            print(f"Job {job_id} failed to start: {error_msg}")
        elif poll_result == 0:
            # Process completed successfully very quickly
            print(f"Job {job_id} completed successfully")
            
        # Save job details (so it can be monitored)
        job_details = {
            "id": job_id,
            "process": process.pid,
            "start_time": time.strftime("%Y-%m-%d %H:%M:%S"),
            "startup_error": poll_result is not None
        }
        with open(os.path.join(job_folder, "job_details.json"), "w") as f:
            json.dump(job_details, f)
            
    except Exception as e:
        # Handle case where subprocess.Popen itself fails
        error_msg = f"Failed to create subprocess: {str(e)}\n{traceback.format_exc()}"
        
        # Write error to log file
        error_log_path = os.path.join(job_folder, "error.log")
        with open(error_log_path, "w") as f:
            f.write(error_msg)
        
        # Write failed status
        status_file = os.path.join(job_folder, "job_status")
        with open(status_file, "w") as f:
            f.write("failed")
        
        print(f"Job {job_id} failed to start: {error_msg}")
        
        # Save job details indicating failure
        job_details = {
            "id": job_id,
            "process": None,
            "start_time": time.strftime("%Y-%m-%d %H:%M:%S"),
            "startup_error": True
        }
        with open(os.path.join(job_folder, "job_details.json"), "w") as f:
            json.dump(job_details, f)



def check_job_status():
    """Check if jobs are still running and return their details"""
    job_data = []
    job_dirs = sorted(os.listdir(JOBS_DIR), reverse=True)

    for job_dir in job_dirs:
        job_details_file = os.path.join(JOBS_DIR, job_dir, "job_details.json")
        job_arguments_file = os.path.join(JOBS_DIR, job_dir, "job_arguments.json")
        if os.path.exists(job_details_file):
            # Load the job details
            with open(job_details_file) as f:
                details = json.load(f)

            # Load current status
            status_file = os.path.join(JOBS_DIR, job_dir, "job_status")
            if os.path.exists(status_file):
                with open(status_file) as f:
                    details["status"] = f.read().strip()
            else:
                # No status file yet, determine status based on process state
                process_id = details.get("process")
                if process_id and is_process_running(process_id):
                    details["status"] = "running"
                else:
                    # Process not running and no status file - check for early exit
                    error_log_file = os.path.join(JOBS_DIR, job_dir, "error.log")
                    if os.path.exists(error_log_file):
                        details["status"] = "failed"
                    elif details.get("startup_error", False):
                        details["status"] = "failed"
                    else:
                        details["status"] = "unknown"

            # Check if process is still running
            process_id = details.get("process")
            details["running"] = is_process_running(process_id) if process_id else False

            # Load error log if it exists
            error_log_file = os.path.join(JOBS_DIR, job_dir, "error.log")
            if os.path.exists(error_log_file):
                with open(error_log_file) as f:
                    details["error_log"] = f.read()

            # Load job arguments
            if os.path.exists(job_arguments_file):
                with open(job_arguments_file) as f:
                    job_arguments = json.load(f)
                    details["arguments"] = job_arguments

            job_data.append(details)

    return job_data

def get_job_status(job_id):
    """Return the job status"""
    status = "running"

    status_file = os.path.join(JOBS_DIR, job_id, "job_status")
    if os.path.exists(status_file):
        with open(status_file) as f:
            status = f.read()

    return status

def cancel_job(job_id):
    """Terminate a running job"""
    job_details_file = os.path.join(JOBS_DIR, job_id, "job_details.json")
    if os.path.exists(job_details_file):
        with open(job_details_file) as f:
            details = json.load(f)

        process_id = details.get("process")
        if os.path.exists(f"/proc/{process_id}"):
            os.kill(process_id, signal.SIGTERM)  # Send termination signal
            details["status"] = "cancelled"
            with open(job_details_file, "w") as f:
                json.dump(details, f)

def remove_job(job_id):
    """Delete job folder"""
    job_folder = os.path.join(JOBS_DIR, job_id)
    if os.path.exists(job_folder):
        for root, dirs, files in os.walk(job_folder, topdown=False):
            for file in files:
                os.remove(os.path.join(root, file))
            for dir in dirs:
                os.rmdir(os.path.join(root, dir))
        os.rmdir(job_folder)

def is_process_running(pid):
    """Check if a process is running"""
    if pid is None:
        return False
    try:
        result = subprocess.run(["ps", "-p", str(pid)], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return result.returncode == 0  # Return True if process is found
    except Exception as e:
        print(f"Error checking process: {e}")
        return False


@app.callback(
    Output("job-status", "children"),
    Input("interval-check", "n_intervals"),
    Input({"type": "cancel-job", "index": ALL}, "n_clicks"),
    Input({"type": "remove-job", "index": ALL}, "n_clicks"),
    State({"type": "cancel-job", "index": ALL}, "id"),
    State({"type": "remove-job", "index": ALL}, "id")
)
def start_or_monitor_job(n_intervals, cancel_clicks, remove_clicks, cancel_ids, remove_ids):
    """Starts a new job, updates job status table, handles job cancellations and removals"""

    triggered_id = ctx.triggered_id

    # Handle job cancellation
    if triggered_id and isinstance(triggered_id, dict) and triggered_id["type"] == "cancel-job":
        cancel_job(triggered_id["index"])

    # Handle job removal
    if triggered_id and isinstance(triggered_id, dict) and triggered_id["type"] == "remove-job":
        remove_job(triggered_id["index"])

    # Monitor job output
    job_data = check_job_status()

    table_header = html.Tr([
        html.Th("Start time"),
        html.Th("Type"),
        html.Th("Algorithm / Model"),
        html.Th("Arguments"),
        html.Th("Status"),
        html.Th("Actions")
    ])

    # On select: download, remove

    table_rows = []
    for job in job_data:
        # Determine job type and algorithm/model
        job_args = job.get("arguments", {})
        job_type = job_args.get("type", "harmonization" if "algorithm" in job_args else "prediction")
        
        # Get algorithm/model from parameters
        if job_type == "harmonization":
            algorithm = job_args.get("algorithm", "Unknown")
        elif job_type == "prediction":
            algorithm = job_args.get("model_name", "Unknown")
        else:
            algorithm = "N/A"
        
        # Create status cell with error details if available
        status_content = job.get("status", "")
        if job.get("error_log") and job.get("status") == "failed":
            # Show a collapsible error details
            status_content = html.Div([
                html.Span("failed", style={"color": "red"}),
                html.Details([
                    html.Summary("Show error details", style={"cursor": "pointer", "marginTop": "5px", "fontSize": "12px", "color": "#6c757d"}),
                    html.Pre(job["error_log"], style={
                        "background": "#f8f9fa",
                        "padding": "10px",
                        "border": "1px solid #dee2e6",
                        "borderRadius": "4px",
                        "maxHeight": "300px",
                        "overflow": "auto",
                        "whiteSpace": "pre-wrap",
                        "fontSize": "12px",
                        "marginTop": "5px",
                        "color": "#212529"
                    })
                ])
            ])
        elif job.get("status") == "failed":
            status_content = html.Span("failed", style={"color": "red"})
        elif job.get("status") == "completed":
            status_content = html.Span("completed", style={"color": "green"})
        elif job.get("running"):
            status_content = html.Span("running", style={"color": "blue"})
        
        # Create actions cell
        actions = []
        if job.get("status", "") in ("completed", "failed"):
            actions.append(html.Button("Download", id={"type": "download-output", "index": job["id"]}, n_clicks=0, style={"margin-right": "5px"}))
        if job.get("running", False):
            actions.append(html.Button("Cancel", id={"type": "cancel-job", "index": job["id"]}, n_clicks=0, style={"margin-right": "5px"}))
        if not job.get("running", False):
            actions.append(html.Button("Remove", id={"type": "remove-job", "index": job["id"]}, n_clicks=0, style={"margin-right": "5px"}))
        
        # Create inputs cell with collapsible arguments
        input_paths = job_args.get("input_paths", [])
        inputs_display = html.Div([
            html.Div(", ".join([os.path.basename(p) for p in input_paths]) if input_paths else "N/A"),
            html.Details([
                html.Summary("Show all arguments", style={"cursor": "pointer", "marginTop": "5px", "fontSize": "12px", "color": "#6c757d"}),
                html.Pre(json.dumps(job_args, indent=2), style={
                    "background": "#f8f9fa",
                    "padding": "10px",
                    "border": "1px solid #dee2e6",
                    "borderRadius": "4px",
                    "maxHeight": "300px",
                    "overflow": "auto",
                    "whiteSpace": "pre-wrap",
                    "fontSize": "12px",
                    "marginTop": "5px",
                    "color": "#212529"
                })
            ])
        ])
        
        table_rows.append(html.Tr([
            html.Td(job.get("start_time", "")),
            html.Td(job_type.capitalize()),
            html.Td(algorithm),
            html.Td(inputs_display),
            html.Td(status_content),
            html.Td(actions)
        ]))

    table_style = {
        "width": "100%",
        "border-collapse": "collapse",
        "font-size": "18px",
        "box-shadow": "0 1px 3px rgba(0,0,0,0.1)"
    }
    
    return html.Table(
        [table_header] + table_rows,
        style=table_style,
        className="job-table"
    )


@app.callback(
    Output("download-data", "data"),
    Input({"type": "download-output", "index": ALL}, "n_clicks"),
    State({"type": "download-output", "index": ALL}, "id"),
    prevent_initial_call=True,
)
def func(n_clicks, ids):
    if not ctx.triggered_id or not isinstance(ctx.triggered_id, dict):
        return dash.no_update  # Prevent unnecessary execution

    triggered_id = ctx.triggered_id
    if triggered_id["type"] == "download-output":

        # Ensure the button was actually clicked this time
        job_id = triggered_id["index"]
        index = [id["index"] for id in ids].index(job_id)
        if n_clicks[index] > 0:
            path = os.path.join(JOBS_DIR, job_id, "output.zip")
            return dcc.send_file(path)

    return dash.no_update  # Avoid unwanted triggers



import os
import sys
import webbrowser
from threading import Timer
from waitress import serve

from dash import html, dcc, Input, Output
import dash_bootstrap_components as dbc

try:
    from importlib.metadata import version
except ImportError:
    from importlib_metadata import version

from cvasl_gui import data_store
from cvasl_gui.app import app
from cvasl_gui.tabs.harmonization import create_tab_harmonization
from cvasl_gui.tabs.prediction import create_tab_prediction
from cvasl_gui.tabs.job_list import create_job_list

data_store.all_data = {
  'harmonization': None,
  'prediction-training': None,
  'prediction-testing': None,
}
data_store.selected_directory = None

app.layout = html.Div(
    id='root',
    children=[html.Div([
        dcc.Tabs(
            id='tabs',
            value='1',
            children=[
                dcc.Tab(
                    label='Harmonize',
                    value='1',
                    style={'backgroundColor': '#f0f0f0'},  # Default tab background
                    selected_style={'backgroundColor': '#e7f1ff', 'fontWeight': 'bold'}  # Active tab style
                ),
                dcc.Tab(
                    label='Predict',
                    value='2',
                    style={'backgroundColor': '#f0f0f0'},  # Default tab background
                    selected_style={'backgroundColor': '#e7f1ff', 'fontWeight': 'bold'}  # Active tab style
                ),
                dcc.Tab(
                    label='Jobs',
                    value='3',
                    style={'backgroundColor': '#f0f0f0'},  # Default tab background
                    selected_style={'backgroundColor': '#e7f1ff', 'fontWeight': 'bold'}  # Active tab style
                ),
            ],
            vertical=False
        ),
        # Load all tab contents here but control visibility through a callback
        html.Div(
            [
                html.Div(create_tab_harmonization(), id='tab-1-content', style={'display': 'none'}),
                html.Div(create_tab_prediction(), id='tab-2-content', style={'display': 'none'}),
                html.Div(create_job_list(), id='tab-3-content', style={'display': 'none'}),
            ],
            id='tab-content-container'
        )
    ], id='main-container')])


# Callback to toggle visibility based on selected tab
@app.callback(
    [Output(f'tab-{i}-content', 'style') for i in range(1, 4)],
    [Input('tabs', 'value')]
)
def display_content(selected_tab):
    # Set 'display' to 'block' for the selected tab and 'none' for others
    return [{'display': 'block' if selected_tab == str(i) else 'none'} for i in range(1, 4)]


def main():
    port = int(os.getenv('CVASL_PORT', 8767))
    # Detect if running as PyInstaller bundle
    is_frozen = getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS')
    # Force production mode for PyInstaller builds
    debug_mode = False if is_frozen else os.getenv('CVASL_DEBUG_MODE', 'False') == 'True'
    path = os.getenv('CVASL_PATHNAME_PREFIX', '/')
    host = '127.0.0.1'

    # Print version information (only in main process, not in reloader)
    if not debug_mode or os.environ.get('WERKZEUG_RUN_MAIN') == 'true':
        try:
            gui_version = version('cvasl-gui')
        except Exception:
            gui_version = "unknown"
        
        try:
            cvasl_version = version('cvasl')
        except Exception:
            cvasl_version = "unknown"
        
        print(f"CVASL GUI version: {gui_version}")
        print(f"CVASL package version: {cvasl_version}")

    app.index_string = """
    <!DOCTYPE html>
    <html>
        <head>
            <title>CVASL</title>
            <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootstrap@5.0.0-beta1/dist/css/bootstrap.min.css">
            <link rel="stylesheet" href="assets/custom.css">
        </head>
        <body>
            {%app_entry%}
            <footer>
                {%config%}
                {%scripts%}
                {%renderer%}
            </footer>
        </body>
    </html>
"""

    # Start the server and open the browser
    if debug_mode:
        app.run_server(port=port, debug=True)
    else:
        # Schedule a timer to open the browser
        url = f"http://{host}:{port}{path}"
        #Timer(1, lambda: webbrowser.open(url)).start()
        
        # Start the server using waitress
        print(f"Starting server at {url}")
        serve(app.server, host=host, port=port, threads=8) 


if __name__ == '__main__':
    main()



import sys
import os
import traceback
import zipfile
import json
import numpy as np
import pandas as pd

from cvasl.prediction import PredictBrainAge
from cvasl.dataset import MRIdataset, encode_cat_features

from sklearn.ensemble import ExtraTreesRegressor, GradientBoostingRegressor
from sklearn.linear_model import LinearRegression, ElasticNetCV
from sklearn.tree import DecisionTreeRegressor
from sklearn.svm import SVR

# Monkey patch to fix the DataFrame ambiguity bug
def _fixed_store_fold_predictions(self, fold_index, y_test, y_pred, y_val, y_pred_val, test_index):
    """Fixed version of _store_fold_predictions that handles DataFrame ambiguity properly."""
    import pandas as pd
    
    # Ensure arrays are properly flattened only if needed
    y_test_flat = y_test.flatten() if hasattr(y_test, 'flatten') else y_test
    y_pred_flat = y_pred.flatten() if hasattr(y_pred, 'flatten') else y_pred
    
    predictions_data = pd.DataFrame({'y_test': y_test_flat, 'y_pred': y_pred_flat})
    predictions_data[self.patient_identifier] = self.data[self.patient_identifier].values[test_index]
    predictions_data['site'] = self.data[self.site_indicator].values[test_index]
    predictions_data['fold'] = fold_index
    
    predictions_data_val = None
    if y_val is not None and y_pred_val is not None and self.data_validation is not None:
        y_val_flat = y_val.flatten() if hasattr(y_val, 'flatten') else y_val
        y_pred_val_flat = y_pred_val.flatten() if hasattr(y_pred_val, 'flatten') else y_pred_val
        
        predictions_data_val = pd.DataFrame({'y_test': y_val_flat, 'y_pred': y_pred_val_flat})
        predictions_data_val[self.patient_identifier] = self.data_validation[self.patient_identifier].values
        predictions_data_val['site'] = self.data_validation[self.site_indicator].values
        predictions_data_val['fold'] = fold_index
    
    return predictions_data, predictions_data_val

# Apply the monkey patch
PredictBrainAge._store_fold_predictions = _fixed_store_fold_predictions


# Argument is the job id (input and parameters(?) are inside the job folder)

WORKING_DIR = os.getenv("CVASL_WORKING_DIRECTORY", ".")
INPUT_DIR = os.path.join(WORKING_DIR, 'data')
JOBS_DIR = os.path.join(WORKING_DIR, 'jobs')

prediction_models = {
  "extratrees": {
    "label": "Extra Trees Regressor",
    "class": ExtraTreesRegressor,
    "parameters": {
        "n_estimators": 100,
        "random_state": np.random.randint(0,100000),
        "criterion": 'absolute_error',
        "min_samples_split": 2,
        "min_samples_leaf": 1,
        "max_features": 'log2',
        "bootstrap": False,
        "n_jobs": -1,
        "warm_start": True
    }
  },
  "gradientboosting": {
    "label": "Gradient Boosting Regressor",
    "class": GradientBoostingRegressor,
    "parameters": {
        "n_estimators": 100,
        "learning_rate": 0.1,
        "max_depth": 3,
        "min_samples_split": 2,
        "min_samples_leaf": 1,
        "subsample": 1.0,
        "random_state": np.random.randint(0, 100000),
        "loss": 'squared_error'
    }
  },
  "linearregression": {
    "label": "Linear Regression",
    "class": LinearRegression,
    "parameters": {
        "fit_intercept": True,
        "n_jobs": -1
    }
  },
  "elasticnetcv": {
    "label": "Elastic Net CV",
    "class": ElasticNetCV,
    "parameters": {
        "l1_ratio": [0.1, 0.5, 0.7, 0.9, 0.95, 0.99, 1],
        "alphas": None,  # Automatically determined
        "cv": 5,
        "max_iter": 1000,
        "tol": 0.0001,
        "random_state": np.random.randint(0, 100000),
        "n_jobs": -1
    }
  },
  "decisiontree": {
    "label": "Decision Tree Regressor", 
    "class": DecisionTreeRegressor,
    "parameters": {
        "max_depth": None,  # No limit
        "min_samples_split": 2,
        "min_samples_leaf": 1,
        "max_features": None,  # Use all features
        "random_state": np.random.randint(0, 100000),
        "splitter": 'best'
    }
  },
  "svr": {
    "label": "Support Vector Regression",
    "class": SVR,
    "parameters": {
        "kernel": 'rbf',
        "C": 1.0,
        "epsilon": 0.1,
        "gamma": 'scale',
        "tol": 0.001,
        "cache_size": 200,
        "max_iter": -1  # No limit
    }
  }
}

def write_job_status(job_id: str, status: str) -> None:
    """ Write the status of the job to a file (for use in the GUI)
    """
    status_path = os.path.join(JOBS_DIR, job_id, "job_status")
    with open(status_path, "w") as f:
        f.write(status)


def zip_job_output(job_id):
    """Create a ZIP file for job output if not already zipped"""
    output_folder = os.path.join(JOBS_DIR, job_id, 'output')
    zip_path = os.path.join(JOBS_DIR, job_id, 'output.zip')

    if not os.path.exists(zip_path):
        with zipfile.ZipFile(zip_path, "w", zipfile.ZIP_DEFLATED) as zip_file:
            for root, _, files in os.walk(output_folder):
                for file in files:
                    file_path = os.path.join(root, file)
                    zip_file.write(file_path, arcname=os.path.relpath(file_path, output_folder))


def run_prediction() -> None:
    """Run the prediction process"""
    
    print("=" * 80)
    print("STARTING PREDICTION JOB")
    print("=" * 80)

    # Load job arguments
    print("\n[1/10] Loading job arguments...")
    job_arguments_path = os.path.join(JOBS_DIR, job_id, "job_arguments.json")
    with open(job_arguments_path) as f:
        job_arguments = json.load(f)
    print(f"   ✓ Job arguments loaded from: {job_arguments_path}")

    model_name = job_arguments["model_name"]
    train_paths = job_arguments["train_paths"]
    train_names = [ os.path.splitext(os.path.basename(path))[0] for path in train_paths ]
    train_sites = job_arguments["train_sites"]
    validation_paths = job_arguments["validation_paths"]
    validation_names = [ os.path.splitext(os.path.basename(path))[0] for path in validation_paths ]
    validation_sites = job_arguments["validation_sites"]
    prediction_features = job_arguments["prediction_features"]
    prediction_features = [x.lower() for x in prediction_features]
    prediction_parameters = job_arguments["parameters"]
    
    label = job_arguments["label"]
    if label is None or label == "":
        label = "predicted"

    print(f"   Model: {model_name}")
    print(f"   Training datasets: {len(train_paths)}")
    print(f"   Validation datasets: {len(validation_paths)}")
    print(f"   Features: {prediction_features}")

    # Load the training datasets into pandas dataframes and concatenate them
    print("\n[2/10] Loading and concatenating training data...")
    train_dfs = [pd.read_csv(path) for path in train_paths]
    train_dfs = pd.concat(train_dfs, ignore_index=True)
    print(f"   ✓ Loaded {len(train_dfs)} training samples")
    
    print("\n[3/10] Loading and concatenating validation data...")
    validation_dfs = [pd.read_csv(path) for path in validation_paths]
    validation_dfs = pd.concat(validation_dfs, ignore_index=True)
    print(f"   ✓ Loaded {len(validation_dfs)} validation samples")

    # Prepare train datasets
    print("\n[4/10] Preparing training datasets...")
    mri_datasets_train = [MRIdataset(input_path, input_site, "participant_id", features_to_drop=[])
                          for input_site, input_path in zip(train_sites, train_paths) ]
    print(f"   ✓ Created {len(mri_datasets_train)} MRIdataset objects")
    
    print("\n[5/10] Preprocessing training datasets...")
    for i, mri_dataset in enumerate(mri_datasets_train):
        print(f"   Processing dataset {i+1}/{len(mri_datasets_train)}...")
        mri_dataset.preprocess()
    print("   ✓ Preprocessing complete")
    
    print("\n[6/10] Encoding categorical features for training data...")
    features_to_map = ['sex']
    encode_cat_features(mri_datasets_train, features_to_map)
    print("   ✓ Categorical features encoded")

    # Prepare test/validation datasets
    print("\n[7/10] Preparing validation datasets...")
    mri_datasets_validation = [MRIdataset(input_path, input_site, "participant_id", features_to_drop=[])
                               for input_site, input_path in zip(validation_sites, validation_paths) ]
    print(f"   ✓ Created {len(mri_datasets_validation)} validation MRIdataset objects")
    
    for i, mri_dataset in enumerate(mri_datasets_validation):
        print(f"   Processing validation dataset {i+1}/{len(mri_datasets_validation)}...")
        mri_dataset.preprocess()
    print("   ✓ Validation preprocessing complete")
    
    features_to_map = ['sex']
    encode_cat_features(mri_datasets_validation, features_to_map)
    print("   ✓ Validation categorical features encoded")

    # Instantiate the model
    print(f"\n[8/10] Instantiating model: {model_name}...")
    model_class = prediction_models.get(model_name).get("class")
    print(f"   Model class: {model_class.__name__}")
    print(f"   Parameters: {prediction_parameters}")
    model = model_class(**prediction_parameters)
    print("   ✓ Model instantiated successfully")

    # Try to avoid the DataFrame ambiguity error by handling validation datasets carefully
    validation_datasets = mri_datasets_validation if mri_datasets_validation else None
    
    # Instantiate the predictor
    print("\n[9/10] Creating PredictBrainAge predictor...")
    predicter = PredictBrainAge(model_name=model_name, model_file_name=model_name, model=model,
                                datasets=mri_datasets_train, datasets_validation=validation_datasets, features=prediction_features,
                                target='age', cat_category='sex', cont_category='age', n_bins=2, splits=1, test_size_p=0.05, random_state=42)
    print("   ✓ Predictor created successfully")

    # Perform training and prediction
    print("\n[10/10] Running prediction (this may take a while)...")
    print("   Training model and making predictions...")
    sys.stdout.flush()  # Force output to be written
    
    metrics_df, metrics_df_val, predictions_df, predictions_df_val, models = predicter.predict()
    
    print("   ✓ Prediction complete!")

    # Some final processing & Write output
    print("\nWriting output files...")
    output_folder = os.path.join(JOBS_DIR, job_id, 'output')
    os.makedirs(output_folder, exist_ok=True)
    print(f"   Output folder: {output_folder}")
    
    # Save metrics
    if metrics_df is not None:
        metrics_path = os.path.join(output_folder, f"metrics_train_{label}.csv")
        metrics_df.to_csv(metrics_path, index=False)
        print(f"   ✓ Saved training metrics: {metrics_path}")
    if metrics_df_val is not None:
        metrics_val_path = os.path.join(output_folder, f"metrics_validation_{label}.csv")
        metrics_df_val.to_csv(metrics_val_path, index=False)
        print(f"   ✓ Saved validation metrics: {metrics_val_path}")
    
    # Process and save predictions
    if predictions_df is not None:
        # Add age_gap calculation if both age columns exist
        df = predictions_df.copy()
        if 'y_pred' in df.columns and 'y_test' in df.columns:
            df['age_gap'] = df['y_pred'] - df['y_test']
            df['age_predicted'] = df['y_pred']
            df['age'] = df['y_test']
        df['label'] = label
        pred_path = os.path.join(output_folder, f"predictions_train_{label}.csv")
        df.to_csv(pred_path, index=False)
        print(f"   ✓ Saved training predictions: {pred_path}")
    
    if predictions_df_val is not None:
        # Add age_gap calculation if both age columns exist
        df = predictions_df_val.copy()
        if 'y_pred' in df.columns and 'y_test' in df.columns:
            df['age_gap'] = df['y_pred'] - df['y_test']
            df['age_predicted'] = df['y_pred']
            df['age'] = df['y_test']
        df['label'] = label
        pred_val_path = os.path.join(output_folder, f"predictions_validation_{label}.csv")
        df.to_csv(pred_val_path, index=False)
        print(f"   ✓ Saved validation predictions: {pred_val_path}")
    
    print("\n" + "=" * 80)
    print("PREDICTION JOB COMPLETED SUCCESSFULLY")
    print("=" * 80)


def get_column_case_insensitive(df, colname):
    match = [c for c in df.columns if c.lower() == colname.lower()]
    if not match:
        raise KeyError(f"Column '{colname}' not found.")
    return df[match[0]]


def process(job_id: str) -> None:
    print(f"\n{'='*80}")
    print(f"INITIALIZING JOB: {job_id}")
    print(f"{'='*80}\n")
    sys.stdout.flush()
    
    write_job_status(job_id, "running")

    try:
        run_prediction()
        
        print("\nCreating output archive...")
        sys.stdout.flush()

        # Zip the output
        zip_job_output(job_id)
        print("✓ Output archive created successfully")
        sys.stdout.flush()

    except KeyboardInterrupt:
        print("\n\nJob interrupted by user")
        sys.stdout.flush()
        write_job_status(job_id, "cancelled")
        with open(os.path.join(JOBS_DIR, job_id, "error.log"), "w") as f:
            f.write("Job cancelled by user\n")
        return
        
    except Exception as e:
        print(f"\n\n{'='*80}")
        print("ERROR: Job failed with exception")
        print(f"{'='*80}")
        print(f"Exception type: {type(e).__name__}")
        print(f"Exception message: {str(e)}")
        print(f"\nFull traceback:")
        print(traceback.format_exc())
        print(f"{'='*80}\n")
        sys.stdout.flush()
        sys.stderr.flush()
        
        write_job_status(job_id, "failed")
        with open(os.path.join(JOBS_DIR, job_id, "error.log"), "w") as f:
            f.write(f"Exception type: {type(e).__name__}\n")
            f.write(f"Exception message: {str(e)}\n\n")
            f.write("Full traceback:\n")
            f.write(traceback.format_exc())
        return
    
    write_job_status(job_id, "completed")
    print(f"\n{'='*80}")
    print(f"JOB COMPLETED: {job_id}")
    print(f"{'='*80}\n")
    sys.stdout.flush()


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print("ERROR: No job ID provided")
        print("Usage: python prediction_job.py <job_id>")
        sys.exit(1)
    
    job_id = sys.argv[1]
    process(job_id)



import time
import dash
from dash import dcc, html, Input, Output, State
import dash_bootstrap_components as dbc
import os
import json
import threading

from cvasl_gui.app import app
from cvasl_gui.tabs.job_list import run_job
from cvasl_gui import data_store
from cvasl_gui.components.directory_input import create_directory_input
from cvasl_gui.components.data_table import create_data_table
from cvasl_gui.components.feature_compare2 import create_feature_compare
from cvasl_gui.tabs.job_list import create_job_list, get_job_status

from cvasl_gui.jobs.prediction_job import prediction_models


# Folder where job output files are stored
WORKING_DIR = os.getenv("CVASL_WORKING_DIRECTORY", ".")
INPUT_DIR = os.path.join(WORKING_DIR, 'data')
JOBS_DIR = os.path.join(WORKING_DIR, 'jobs')


def create_tab_prediction():
    return html.Div([
        dbc.Accordion([
            dbc.AccordionItem([create_directory_input('prediction-training')],
                title="Select training data"),
            dbc.AccordionItem([create_data_table('prediction-training')],
                title="Inspect training data"),
            dbc.AccordionItem([create_directory_input('prediction-testing')],
                title="Select testing data"),
            dbc.AccordionItem([create_data_table('prediction-testing')],
                title="Inspect testing data"),
            dbc.AccordionItem([create_feature_compare()],
                title="Feature comparison"),
            dbc.AccordionItem(create_prediction_parameters(),
                title="Prediction"),
            dcc.Store(id="prediction-job-id", data=None),
        ], always_open=True)
    ])


def get_dataframe_columns():
    df = data_store.all_data['prediction-training']
    if df is None:
        return []
    return df.columns


def create_prediction_parameters():
    return [
        # Row for model selection
        dbc.Row([
            dbc.Col(html.Label("Model:", style={"marginTop": "6px"}), width=3),
            dbc.Col(
                dcc.Dropdown(
                    id="model-dropdown",
                    options=[{"label": v["label"], "value": k} for k, v in prediction_models.items()],
                    value="extratrees",
                    clearable=False,
                ),
            ),
        ], className="mb-3"),

        # Row for main feature selection
        dbc.Row([
            dbc.Col(html.Label("Features:", style={"marginTop": "6px"}), width=3),
            dbc.Col(
                dcc.Dropdown(
                    id="prediction-features-dropdown",
                    options=[{"label": col, "value": col} for col in get_dataframe_columns()],
                    multi=True,
                    placeholder="Select features...",
                ),
            ),
        ], className="mb-3"),

        # Row for label text
        dbc.Row([
            dbc.Col(html.Label("Label:", style={"marginTop": "6px"}), width=3),
            dbc.Col(
                dbc.Input(
                    id="prediction-label-input",
                    type="text",
                    placeholder="Enter label...",
                    value="predicted",
                ),
            ),
        ], className="mb-3"),

        # Row for prediction button and status
        dbc.Row([
            dbc.Col(html.Label("", style={"marginTop": "6px"}), width=3),
            dbc.Col(
                html.Div([
                    html.Button("Run prediction", id="prediction-start-button", n_clicks=0),
                    html.Span("Status: ", style={"marginLeft": "10px"}),
                    html.Span(id="prediction-job-status", children="Idle"),
                    dcc.Interval(id="prediction-status-interval", interval=1000, n_intervals=0, disabled=True)
                ]),
            ),
        ], className="mb-3"),
    ]


# Populate dropdown with columns from the data table
@app.callback(
    Output("prediction-features-dropdown", "options"),
    Output("prediction-features-dropdown", "value"),
    Input({'type': 'data-table', 'index': 'prediction-training'}, "data"),
    prevent_initial_call=True
)
def update_feature_dropdown(data):
    if not data:
        return [], []
    options = [{"label": col, "value": col} for col in data[0].keys()]
    look_for_values = ['aca_b_cbf', 'aca_b_cov', 'csf_vol', 'gm_icvratio', 'gm_vol', 'gmwm_icvratio',
                       'mca_b_cbf', 'mca_b_cov','pca_b_cbf', 'pca_b_cov', 'totalgm_b_cbf','totalgm_b_cov',
                       'wm_vol', 'wmh_count', 'wmhvol_wmvol']
    default_values = [col for col in data[0].keys() if col.lower() in look_for_values]
    return options, default_values


@app.callback(
    Output("prediction-job-id", "data"),
    Output("prediction-status-interval", "disabled", allow_duplicate=True),
    Input("prediction-start-button", "n_clicks"),
    State("model-dropdown", "value"),
    State("prediction-features-dropdown", "value"),
    State("prediction-label-input", "value"),
    prevent_initial_call=True,
)
def start_job(n_clicks, model_name, selected_features, label):
    if not selected_features:
        return dash.no_update, True

    job_arguments = {
        "type": "prediction",
        "model_name": model_name,
        "train_paths": data_store.input_files['prediction-training'],
        "train_sites": data_store.input_sites['prediction-training'],
        "validation_paths": data_store.input_files['prediction-testing'],
        "validation_sites": data_store.input_sites['prediction-testing'],
        "prediction_features": selected_features,
        "parameters": {
            **prediction_models[model_name]["parameters"]
        },
        "label": label,
    }

    # Generate job_id and start the job in a separate thread
    job_id = str(int(time.time())) + "_" + model_name
    threading.Thread(target=run_job, args=(job_arguments, job_id, False), daemon=True).start()

    return job_id, False  # enable the interval


@app.callback(
    Output("prediction-job-status", "children"),
    Output("prediction-status-interval", "disabled", allow_duplicate=True),
    Input("prediction-status-interval", "n_intervals"),
    State("prediction-job-id", "data"),
    prevent_initial_call=True,
)
def update_job_status(n, job_id):
    if not job_id:
        return "", True

    status = get_job_status(job_id)

    if status.lower() in ("completed", "failed", "cancelled"):
        return status, True  # Stop interval
    return status, False



parameters = {
    "features_to_harmonize": {
        "type": "feature-list-multi",
        "label": "Features to harmonize",
        "description": "Select the features to be harmonised across sites.",
    },
    "batch_list_harmonisations": {
        "type": "feature-list-multi",
        "label": "Batch list harmonisations",
        "description": "Select column names indicating batch effect within covars (i.e. ['PLD', 'LD',’readout’]).",
    },
    "covariates": {
        "type": "feature-list-multi",
        "label": "Covariates",
        "description": "Select the covariates to be protected during harmonisation across sites.",
    },
    "discrete_covariates": {
        "type": "feature-list-multi",
        "label": "Discrete covariates",
        "description": "Select the discrete/categorical covariates to be protected during harmonisation across sites, such as sex.",
    },
    "continuous_covariates": {
        "type": "feature-list-multi",
        "label": "Continuous covariates",
        "description": "Select the continuous covariates to be protected during harmonisation across sites, such as age.",
    },
    "numerical_covariates": {
        "type": "feature-list-multi",
        "label": "Numerical covariates",
        "description": "Determine which covariates are numerical, others will be viewed as categorical.",
    },
    "smooth_terms": {
        "type": "feature-list-multi",
        "label": "Smooth terms",
        "description": "Select the covariates that are assumed to be non-linearly associated with the features.",
    },
    "discrete_cluster_features": {
        "type": "feature-list-multi",
        "label": "Discrete cluster features",
        "description": "Select the discrete/categorical cluster features used to determine site indicators, such as 'readout'.",
    },
    "continuous_cluster_features": {
        "type": "feature-list-multi",
        "label": "Continuous cluster features",
        "description": "Select the continuous cluster features used to determine site indicators, such as ‘PLD’, ‘LD’.",
    },
    "metric": {
        "type": "selection",
        "label": "Metric",
        "description": "Metric to define the optimal number of clusters.",
        "options": ["distortion", "silhouette", "calinski_harabasz"],
        "hide": True,
    },
    "features_reduction": {
        "type": "selection",
        "label": "Features reduction",
        "description": "Method for reducing the embedded space with n components. Can be None, 'pca' or 'umap'.",
        "options": ["None", "pca", "umap"],
    },
    "feature_reduction_dimensions": {
        "type": "int",
        "label": "Feature reduction dimensions",
        "description": "Dimension of the embedded space for features reduction.",
        "default": 2,
        "hide": True,
    },
    "empirical_bayes": {
        "type": "bool",
        "label": "Empirical bayes",
        "description": "Whether to use empirical Bayes (EB) estimates of site effects.",
        "default": True,
    },
    "mean_only": {
        "type": "bool",
        "label": "Mean only",
        "description": "Select if harmonisation of only the mean, instead of both the mean and variance (scaling), is performed.",
        "default": False,
    },
    "parametric": {
        "type": "bool",
        "label": "Parametric",
        "description": "Deselect if non-parametric adjustments should be performed.",
        "default": True,
    },
    "use_gmm": {
        "type": "bool",
        "label": "Use GMM",
        "description": "Select to use Gaussian Mixture Modeling (GMM) to identify scan groupings associated with hidden covariates to better address bimodal feature distributions. Off by default.",
        "default": False,
    },
    "site_indicator": {
        "type": "feature-list-single",
        "label": "Site indicator",
        "description": "The site identifier.",
        "default": "site",
    },
    "patient_identifier": {
        "type": "feature-list-single",
        "label": "Patient identifier",
        "description": "Column name identifying each patient.",
        "default": "participant_id",
        "hide": True,
    },
    "intermediate_results_path": {
        "type": "str",
        "label": "Intermediate results path",
        "description": "Path to save intermediate results.",
        "default": ".",
        "hide": True,
    },
    "return_extended": {
        "type": "bool",
        "label": "Return extended",
        "description": "Whether to return extended outputs (intermediate dataframes).",
        "default": False,
        "hide": True,
    },
    "data_subset": {
        "type": "feature-list-multi",
        "label": "Data subset",
        "description": "Features of the dataset subset to be passed to autocombat for harmonization.",
        "hide": True,
    },
}



import json
import os
from dash import dcc, html, Input, Output, State, MATCH, ctx
from dash.exceptions import PreventUpdate
import pandas as pd
from cvasl_gui.app import app
from cvasl_gui import data_store


def create_directory_input(instance_id):
    return html.Div([
        html.Div(id={'type': 'file-list-container', 'index': instance_id}, children=[
            dcc.Checklist(
                id={'type': 'file-list', 'index': instance_id},
                options=[],
                labelStyle={'display': 'block'},
                inputStyle={'marginRight': '5px'},
                inline=True
            )
        ]),
        html.Div(id={'type': 'harmonized-file-list-container', 'index': instance_id}, children=[
            dcc.Checklist(
                id={'type': 'harmonized-file-list', 'index': instance_id},
                options=[],
                labelStyle={'display': 'block'},
                inputStyle={'marginRight': '5px'},
                inline=True
            )
        ]),
        html.Div(id={'type': 'file-contents-container', 'index': instance_id}),
    ], style={'display': 'flex', 'flex-direction': 'column', 'gap': '10px'})


WORKING_DIR = os.getenv("CVASL_WORKING_DIRECTORY", ".")
INPUT_DIR = os.path.join(WORKING_DIR, 'data')
JOBS_DIR = os.path.join(WORKING_DIR, 'jobs')

@app.callback(
    Output({'type': 'file-list', 'index': MATCH}, 'options'),
    Input({'type': 'file-list', 'index': MATCH}, 'id')  # dummy
)
def populate_file_list(_):
    if not os.path.isdir(INPUT_DIR):
        return [{'label': 'Directory not found', 'value': '', 'disabled': True}]
    files = sorted([
        f for f in os.listdir(INPUT_DIR)
        if os.path.isfile(os.path.join(INPUT_DIR, f)) and f.endswith('.csv')
    ])
    return [{'label': f, 'value': f} for f in files]

@app.callback(
    Output({'type': 'harmonized-file-list', 'index': MATCH}, 'options'),
    Input({'type': 'harmonized-file-list', 'index': MATCH}, 'id')  # dummy
)
def populate_harmonized_file_list(_):
    if not os.path.isdir(JOBS_DIR):
        return [{'label': 'Directory not found', 'value': '', 'disabled': True}]
    job_dirs = sorted(os.listdir(JOBS_DIR), reverse=True)
    all_files = []
    for job_dir in job_dirs:
        job_details_file = os.path.join(JOBS_DIR, job_dir, "job_details.json")
        job_arguments_file = os.path.join(JOBS_DIR, job_dir, "job_arguments.json")
        output_dir = os.path.join(JOBS_DIR, job_dir, 'output')
        if os.path.exists(output_dir):
            output_files = [ os.path.join(job_dir, f) for f in os.listdir(output_dir)
                            if os.path.isfile(os.path.join(output_dir, f)) and f.endswith('.csv') ]
            all_files.extend(output_files)

    return [{'label': f, 'value': f} for f in all_files]
    
        # if os.path.exists(job_details_file):
        #     # Load the job details
        #     with open(job_details_file) as f:
        #         details = json.load(f)

        #     # Load current status
        #     status_file = os.path.join(JOBS_DIR, job_dir, "job_status")
        #     if os.path.exists(status_file):
        #         with open(status_file) as f:
        #             details["status"] = f.read()

        #     # Check if process is still running
        #     process_id = details.get("process")
        #     details["running"] = is_process_running(process_id)

        #     # Load job arguments
        #     if os.path.exists(job_arguments_file):
        #         with open(job_arguments_file) as f:
        #             job_arguments = json.load(f)
        #             details["arguments"] = job_arguments

        #     job_data.append(details)


@app.callback(
    Output({'type': 'file-contents-container', 'index': MATCH}, 'children'),
    Input({'type': 'file-list', 'index': MATCH}, 'value'),
    Input({'type': 'harmonized-file-list', 'index': MATCH}, 'value')
)
def load_all_selected_files(normal_files, harmonized_files):
    trigger_id = ctx.triggered_id
    if isinstance(trigger_id, dict) and 'index' in trigger_id:
        ctx_index = trigger_id['index']
    else:
        raise PreventUpdate

    # If nothing selected, clear data
    if not normal_files and not harmonized_files:
        data_store.input_files[ctx_index] = []
        data_store.input_sites[ctx_index] = []
        data_store.all_data[ctx_index] = pd.DataFrame()
        return html.Div("No files selected")

    normal_dfs = []
    harmonized_dfs = []
    normal_file_rows = []
    harmonized_file_rows = []
    errors = []

    # Load normal files
    input_files = []
    input_sites = []
    if normal_files:
        for fname in normal_files:
            file_path = os.path.join(INPUT_DIR, fname)
            input_files.append(file_path)
            try:
                df = pd.read_csv(file_path)
                # Get the site from the dataframe
                site_col = [col for col in df.columns if col.lower() == 'site']
                if site_col:
                    input_sites.append(int(df[site_col[0]].iloc[0]))
                else:
                    input_sites.append(0) #TODO: how to make sure this is sensible?
                if not 'label' in df.columns:
                    df['label'] = 'original'
                normal_dfs.append(df)
                normal_file_rows.append(f"{fname} ({len(df)})")
            except Exception as e:
                errors.append(f"Error loading {fname}: {e}")
        data_store.input_files[ctx_index] = input_files
        data_store.input_sites[ctx_index] = input_sites
        #data_store.input_data = pd.concat(normal_dfs, ignore_index=True) if normal_dfs else pd.DataFrame()

    # Load harmonized files
    if harmonized_files:
        for fname in harmonized_files:
            split = fname.split(os.sep)
            file_path = os.path.join(JOBS_DIR, split[0], 'output', split[1])
            try:
                df = pd.read_csv(file_path)
                if not 'label' in df.columns:
                    df['label'] = 'harmonized'
                harmonized_dfs.append(df)
                harmonized_file_rows.append(f"{fname} ({len(df)})")
            except Exception as e:
                errors.append(f"Error loading {fname}: {e}")

    # Combine all
    combined_dfs = normal_dfs + harmonized_dfs
    data_store.all_data[ctx_index] = pd.concat(combined_dfs, ignore_index=True) if combined_dfs else pd.DataFrame()

    return html.Div([
        # html.Div(f"Loaded normal files: {', '.join(normal_file_rows) if normal_file_rows else 'None'}"),
        # html.Div(f"Loaded harmonized files: {', '.join(harmonized_file_rows) if harmonized_file_rows else 'None'}"),
        html.Div(f"{len(data_store.all_data[ctx_index])} total rows loaded"),
        *([html.Div(e, style={'color': 'red'}) for e in errors] if errors else [])
    ])



import pandas as pd
from dash import Input, Output, dash_table, MATCH, ctx
from dash.exceptions import PreventUpdate
from cvasl_gui.app import app
from cvasl_gui import data_store


def create_data_table(instance_id):
    return dash_table.DataTable(
        id={'type': 'data-table', 'index': instance_id},
        columns=[
            {'name': 'participant_id', 'id': 'participant_id', 'type': 'text'},
            {'name': 'ID', 'id': 'ID', 'type': 'text'},
            {'name': 'Age', 'id': 'Age', 'type': 'numeric'},
            {'name': 'Sex', 'id': 'Sex', 'type': 'text'},
            {'name': 'Site', 'id': 'Site', 'type': 'text'}
        ],
        data=[],
        filter_action='native',
        sort_action='native',
        sort_mode='multi',
        page_action='native',
        page_size=20,
        style_table={'overflowX': 'auto'}
    )

@app.callback(
    Output({'type': 'data-table', 'index': MATCH}, 'data'),
    Output({'type': 'data-table', 'index': MATCH}, 'columns'),
    Input({'type': 'file-contents-container', 'index': MATCH}, 'children')
)
def update_table(data):
    if data is None:
        raise PreventUpdate
    if not ctx.triggered_id or not isinstance(ctx.triggered_id, dict):
        raise PreventUpdate

    ctx_index = ctx.triggered_id['index']
    df = data_store.all_data[ctx_index]
    columns = [{'name': col, 'id': col} for col in df.columns]
    return df.to_dict('records'), columns



from dash import Dash
import dash_bootstrap_components as dbc
import os
from dotenv import load_dotenv

# Load environment variables
load_dotenv()

icons = "https://cdn.jsdelivr.net/npm/bootstrap-icons@1.10.5/font/bootstrap-icons.css"

external_stylesheets = [dbc.themes.BOOTSTRAP, icons, "assets/custom.css"]

pathname_prefix = os.getenv("CVASL_PATHNAME_PREFIX", "/")
print(f"Using pathname prefix: {pathname_prefix}")

app = Dash(
    __name__,
    requests_pathname_prefix=pathname_prefix,
    routes_pathname_prefix=pathname_prefix,
    external_stylesheets=external_stylesheets)



import sys
import os
import traceback
import zipfile
import json

from cvasl.harmonizers import NeuroCombat, NeuroHarmonize, Covbat, NestedComBat, \
                              AutoCombat, RELIEF
from cvasl.dataset import MRIdataset, encode_cat_features


# Argument is the job id (input and parameters(?) are inside the job folder)

WORKING_DIR = os.getenv("CVASL_WORKING_DIRECTORY", ".")
INPUT_DIR = os.path.join(WORKING_DIR, 'data')
JOBS_DIR = os.path.join(WORKING_DIR, 'jobs')

harmonizers = {
    "neurocombat": NeuroCombat,
    "neuroharmonize": NeuroHarmonize,
    "covbat": Covbat,
    "nestedcombat": NestedComBat,
    "autocombat": AutoCombat,
    "relief": RELIEF
}

def write_job_status(job_id: str, status: str) -> None:
    """ Write the status of the job to a file (for use in the GUI)
    """
    status_path = os.path.join(JOBS_DIR, job_id, "job_status")
    with open(status_path, "w") as f:
        f.write(status)


def zip_job_output(job_id):
    """Create a ZIP file for job output if not already zipped"""
    output_folder = os.path.join(JOBS_DIR, job_id, 'output')
    zip_path = os.path.join(JOBS_DIR, job_id, 'output.zip')

    if not os.path.exists(zip_path):
        with zipfile.ZipFile(zip_path, "w", zipfile.ZIP_DEFLATED) as zip_file:
            for root, _, files in os.walk(output_folder):
                for file in files:
                    file_path = os.path.join(root, file)
                    zip_file.write(file_path, arcname=os.path.relpath(file_path, output_folder))


def run_harmonization() -> None:
    """Run the harmonization process"""

    # Load job arguments
    job_arguments_path = os.path.join(JOBS_DIR, job_id, "job_arguments.json")
    with open(job_arguments_path) as f:
        job_arguments = json.load(f)

    input_paths = job_arguments["input_paths"]
    input_names = [ os.path.splitext(os.path.basename(path))[0] for path in input_paths ]
    input_sites = job_arguments["input_sites"]

    harmonization_parameters = job_arguments["parameters"]

    label = job_arguments["label"]
    if label is None or label == "":
        label = "harmonized"

    print("Running harmonization")
    print("Input paths:", input_paths)
    print("Parameters:", harmonization_parameters)

    # Create the datasets
    mri_datasets = [MRIdataset(input_path, input_site, "participant_id", features_to_drop=[])
                    for input_site, input_path in zip(input_sites, input_paths) ]
    for mri_dataset in mri_datasets:
        mri_dataset.preprocess()
    features_to_map = ['sex']
    encode_cat_features(mri_datasets, features_to_map)

    # Instantiate the correct harmonizer
    harmonizer_class = harmonizers.get(job_arguments["algorithm"])
    harmonizer = harmonizer_class(**harmonization_parameters)

    # Perform the harmonization
    output = harmonizer.harmonize(mri_datasets)

    # Prepare the output datasets and add the label column
    for dataset in output:
        dataset.prepare_for_export()
    dfs = [dataset.data for dataset in output]
    for df in dfs:
        df['label'] = label

    # Write output
    output_folder = os.path.join(JOBS_DIR, job_id, 'output')
    os.makedirs(output_folder, exist_ok=True)
    for i, df in enumerate(dfs):
        df.to_csv(os.path.join(output_folder, f"{input_names[i]}_{label}.csv"), index=False)


def process(job_id: str) -> None:
    write_job_status(job_id, "running")
    print("Processing job", job_id)

    try:
        run_harmonization()

        # Zip the output
        zip_job_output(job_id)

    except Exception as e:
        write_job_status(job_id, "failed")
        with open(os.path.join(JOBS_DIR, job_id, "error.log"), "w") as f:
            f.write(traceback.format_exc())
        return
    
    write_job_status(job_id, "completed")


if __name__ == '__main__':
    job_id = sys.argv[1]
    process(job_id)



from dash import html, dcc, Input, Output
import pandas as pd
from cvasl_gui.app import app
import plotly.express as px
from cvasl_gui import data_store


layout = html.Div([
    html.Div([
        html.Label("X-axis"),
        dcc.Dropdown(
            id='scatter-x-axis2',
            placeholder="Select a column"
        ),
        html.Label("Y-axis"),
        dcc.Dropdown(
            id='scatter-y-axis2',
            placeholder="Select a column"
        ),
        html.Label("Group by"),
        dcc.Dropdown(
            id='scatter-group-by2',
            placeholder="Select a column"
        )
    ], style={'flex': '1', 'padding': '10px'}),
    
    html.Div([
      dcc.Graph(id='scatter-plot2')
    ], style={'flex': '2', 'padding': '10px'})
], style={'display': 'flex', 'width': '100%'})


@app.callback(
    Output('scatter-x-axis2', 'options'),
    Output('scatter-x-axis2', 'value'),
    Output('scatter-y-axis2', 'options'),
    Output('scatter-y-axis2', 'value'),
    Output('scatter-group-by2', 'options'),
    Output('scatter-group-by2', 'value'),
    Input({'type': 'data-table', 'index': 'prediction-training'}, 'data'),
    Input({'type': 'data-table', 'index': 'prediction-testing'}, 'data'),
)
def update_scatter_dropdowns(data, data2):
    df = data_store.all_data['prediction-training']
    if not hasattr(data_store, 'all_data') or df is None:
        return [], None, [], None, [], None

    columns = [{'label': col, 'value': col} for col in df.columns]
    default_x_column = 'Age' if 'Age' in df.columns else None
    default_y_column = df.columns[5] if len(df.columns) > 5 else None
    default_group_column = 'Site' if 'Site' in df.columns else None
    return columns, default_x_column, columns, default_y_column, columns, default_group_column


@app.callback(
    Output('scatter-plot2', 'figure'),
    Input('scatter-x-axis2', 'value'),
    Input('scatter-y-axis2', 'value'),
    Input('scatter-group-by2', 'value')
)
def update_scatter_plot(x_axis, y_axis, group_by):
    if not x_axis or not y_axis:
        return {}
    df1 = data_store.all_data['prediction-training']
    df2 = data_store.all_data['prediction-testing']
    data = pd.concat([df1, df2], ignore_index=True)
    fig = px.scatter(data, x=x_axis, y=y_axis, color=group_by, facet_col='label', trendline='ols')
    return fig
